@@ -3,16 +3,17 @@
     tests.test_sync
     ~~~~~~~~~~~~~~~
 
-    :copyright: (c) 2014 Markus Unterwaditzer
+    :copyright: (c) 2014 Markus Unterwaditzer & contributors
     :license: MIT, see LICENSE for more details.
 '''
 
 import pytest
+
 from vdirsyncer.storage.base import Item
 from vdirsyncer.storage.memory import MemoryStorage
-from vdirsyncer.sync import sync
+from vdirsyncer.sync import BothReadOnly, StorageEmpty, SyncConflict, sync
+
 from . import assert_item_equals, normalize_item
-import vdirsyncer.exceptions as exceptions
 
 
 def empty_storage(x):
@@ -22,7 +23,7 @@ def empty_storage(x):
 def test_irrelevant_status():
     a = MemoryStorage()
     b = MemoryStorage()
-    status = {'1': ('1.txt', 1234, '1.ics', 2345)}
+    status = {'1': ('1', 1234, '1.ics', 2345)}
     sync(a, b, status)
     assert not status
     assert empty_storage(a)
@@ -33,13 +34,13 @@ def test_missing_status():
     a = MemoryStorage()
     b = MemoryStorage()
     status = {}
-    item = Item(u'UID:1')
-    a.upload(item)
-    b.upload(item)
+    item = Item(u'asdf')
+    href_a, _ = a.upload(item)
+    href_b, _ = b.upload(item)
     sync(a, b, status)
     assert len(status) == 1
-    assert a.has('1.txt')
-    assert b.has('1.txt')
+    assert a.has(href_a)
+    assert b.has(href_b)
 
 
 def test_missing_status_and_different_items():
@@ -50,12 +51,12 @@ def test_missing_status_and_different_items():
     item2 = Item(u'UID:1\nhoho')
     a.upload(item1)
     b.upload(item2)
-    with pytest.raises(exceptions.SyncConflict):
+    with pytest.raises(SyncConflict):
         sync(a, b, status)
     assert not status
     sync(a, b, status, conflict_resolution='a wins')
-    assert_item_equals(item1, b.get('1.txt')[0])
-    assert_item_equals(item1, a.get('1.txt')[0])
+    assert_item_equals(item1, b.get('1')[0])
+    assert_item_equals(item1, a.get('1')[0])
 
 
 def test_upload_and_update():
@@ -66,22 +67,22 @@ def test_upload_and_update():
     item = Item(u'UID:1')  # new item 1 in a
     a.upload(item)
     sync(a, b, status)
-    assert_item_equals(b.get('1.txt')[0], item)
+    assert_item_equals(b.get('1')[0], item)
 
     item = Item(u'UID:1\nASDF:YES')  # update of item 1 in b
-    b.update('1.txt', item, b.get('1.txt')[1])
+    b.update('1', item, b.get('1')[1])
     sync(a, b, status)
-    assert_item_equals(a.get('1.txt')[0], item)
+    assert_item_equals(a.get('1')[0], item)
 
     item2 = Item(u'UID:2')  # new item 2 in b
     b.upload(item2)
     sync(a, b, status)
-    assert_item_equals(a.get('2.txt')[0], item2)
+    assert_item_equals(a.get('2')[0], item2)
 
     item2 = Item(u'UID:2\nASDF:YES')  # update of item 2 in a
-    a.update('2.txt', item2, a.get('2.txt')[1])
+    a.update('2', item2, a.get('2')[1])
     sync(a, b, status)
-    assert_item_equals(b.get('2.txt')[0], item2)
+    assert_item_equals(b.get('2')[0], item2)
 
 
 def test_deletion():
@@ -91,17 +92,18 @@ def test_deletion():
 
     item = Item(u'UID:1')
     a.upload(item)
+    a.upload(Item(u'UID:2'))
     sync(a, b, status)
-    b.delete('1.txt', b.get('1.txt')[1])
+    b.delete('1', b.get('1')[1])
     sync(a, b, status)
-    assert not a.has('1.txt') and not b.has('1.txt')
+    assert not a.has('1') and not b.has('1')
 
     a.upload(item)
     sync(a, b, status)
-    assert a.has('1.txt') and b.has('1.txt')
-    a.delete('1.txt', a.get('1.txt')[1])
+    assert a.has('1') and b.has('1')
+    a.delete('1', a.get('1')[1])
     sync(a, b, status)
-    assert not a.has('1.txt') and not b.has('1.txt')
+    assert not a.has('1') and not b.has('1')
 
 
 def test_already_synced():
@@ -111,7 +113,7 @@ def test_already_synced():
     a.upload(item)
     b.upload(item)
     status = {
-        '1': ('1.txt', a.get('1.txt')[1], '1.txt', b.get('1.txt')[1])
+        '1': ('1', a.get('1')[1], '1', b.get('1')[1])
     }
     old_status = dict(status)
     a.update = b.update = a.upload = b.upload = \
@@ -120,7 +122,7 @@ def test_already_synced():
     for i in (1, 2):
         sync(a, b, status)
         assert status == old_status
-        assert a.has('1.txt') and b.has('1.txt')
+        assert a.has('1') and b.has('1')
 
 
 @pytest.mark.parametrize('winning_storage', 'ab')
@@ -135,7 +137,7 @@ def test_conflict_resolution_both_etags_new(winning_storage):
     assert status
     a.update(href_a, Item(u'UID:1\nitem a'), etag_a)
     b.update(href_b, Item(u'UID:1\nitem b'), etag_b)
-    with pytest.raises(exceptions.SyncConflict):
+    with pytest.raises(SyncConflict):
         sync(a, b, status)
     sync(a, b, status, conflict_resolution='{} wins'.format(winning_storage))
     item_a, _ = a.get(href_a)
@@ -146,6 +148,21 @@ def test_conflict_resolution_both_etags_new(winning_storage):
     assert u'item {}'.format(winning_storage) in n
 
 
+def test_updated_and_deleted():
+    a = MemoryStorage()
+    b = MemoryStorage()
+    href_a, etag_a = a.upload(Item(u'UID:1'))
+    status = {}
+    sync(a, b, status, force_delete=True)
+
+    (href_b, etag_b), = b.list()
+    b.delete(href_b, etag_b)
+    a.update(href_a, Item(u'UID:1\nupdated'), etag_a)
+    sync(a, b, status, force_delete=True)
+
+    assert len(list(a.list())) == len(list(b.list())) == 1
+
+
 def test_conflict_resolution_invalid_mode():
     a = MemoryStorage()
     b = MemoryStorage()
@@ -192,3 +209,53 @@ def test_uses_get_multi(monkeypatch):
 
     sync(a, b, {})
     assert get_multi_calls == [[expected_href]]
+
+
+def test_empty_storage_dataloss():
+    a = MemoryStorage()
+    b = MemoryStorage()
+    a.upload(Item(u'UID:1'))
+    a.upload(Item(u'UID:2'))
+    status = {}
+    sync(a, b, status)
+    with pytest.raises(StorageEmpty):
+        sync(MemoryStorage(), b, status)
+
+    with pytest.raises(StorageEmpty):
+        sync(a, MemoryStorage(), status)
+
+
+def test_no_uids():
+    a = MemoryStorage()
+    b = MemoryStorage()
+    href_a, _ = a.upload(Item(u'ASDF'))
+    href_b, _ = b.upload(Item(u'FOOBAR'))
+    status = {}
+    sync(a, b, status)
+
+    a_items = set(a.get(href)[0].raw for href, etag in a.list())
+    b_items = set(b.get(href)[0].raw for href, etag in b.list())
+
+    assert a_items == b_items == {u'ASDF', u'FOOBAR'}
+
+
+def test_both_readonly():
+    a = MemoryStorage(read_only=True)
+    b = MemoryStorage(read_only=True)
+    assert a.read_only
+    assert b.read_only
+    status = {}
+    with pytest.raises(BothReadOnly):
+        sync(a, b, status)
+
+
+def test_readonly():
+    a = MemoryStorage()
+    b = MemoryStorage(read_only=True)
+    status = {}
+    href_a, _ = a.upload(Item(u'UID:1'))
+    href_b, _ = b.upload(Item(u'UID:2'))
+    sync(a, b, status)
+    assert len(status) == 2 and a.has(href_a) and not b.has(href_a)
+    sync(a, b, status)
+    assert len(status) == 1 and not a.has(href_a) and not b.has(href_a)
