@@ -9,4 +9,7 @@
     :copyright: (c) 2014 Markus Unterwaditzer & contributors
     :license: MIT, see LICENSE for more details.
 '''
-__version__ = '0.3.0'
+__version__ = '0.3.1'
+
+PROJECT_HOME = 'https://github.com/untitaker/vdirsyncer'
+DOCS_HOME = 'https://vdirsyncer.readthedocs.org/en/latest'
